attribute vec4 tangent;

uniform vec2 uOffset;
uniform vec2 uRepeat;

uniform bool enableDisplacement;

#ifdef VERTEX_TEXTURES
	uniform sampler2D tDisplacement;
	uniform float uDisplacementScale;
	uniform float uDisplacementBias;
#endif

varying vec3 vTangent;
varying vec3 vBinormal;
varying vec3 vNormal;
varying vec2 vUv;

varying vec3 vWorldPosition;
varying vec3 vViewPosition;

//==============================================
// THREE.ShaderChunk[ "skinning_pars_vertex" ],
//==============================================
#ifdef USE_SHADOWMAP
	uniform sampler2D shadowMap[ MAX_SHADOWS ];
	uniform vec2 shadowMapSize[ MAX_SHADOWS ];

	uniform float shadowDarkness[ MAX_SHADOWS ];
	uniform float shadowBias[ MAX_SHADOWS ];

	varying vec4 vShadowCoord[ MAX_SHADOWS ];

	float unpackDepth( const in vec4 rgba_depth ) {
		const vec4 bit_shift = vec4( 1.0 / ( 256.0 * 256.0 * 256.0 ), 1.0 / ( 256.0 * 256.0 ), 1.0 / 256.0, 1.0 );
		float depth = dot( rgba_depth, bit_shift );
		return depth;
	}
#endif
//==============================================
//==============================================
//==============================================


//==============================================
// THREE.ShaderChunk[ "shadowmap_pars_vertex" ],
//==============================================
#ifdef USE_SHADOWMAP
	varying vec4 vShadowCoord[ MAX_SHADOWS ];
	uniform mat4 shadowMatrix[ MAX_SHADOWS ];
#endif
//==============================================
//==============================================
//==============================================


//==============================================
// THREE.ShaderChunk[ "logdepthbuf_pars_vertex" ],
//==============================================
#ifdef USE_LOGDEPTHBUF
	#ifdef USE_LOGDEPTHBUF_EXT
		varying float vFragDepth;
	#endif
	uniform float logDepthBufFC;
#endif
//==============================================
//==============================================
//==============================================




void main() {

	//==============================================
	// THREE.ShaderChunk[ "skinbase_vertex" ],
	//==============================================
	#ifdef USE_SKINNING
		mat4 boneMatX = getBoneMatrix( skinIndex.x );
		mat4 boneMatY = getBoneMatrix( skinIndex.y );
		mat4 boneMatZ = getBoneMatrix( skinIndex.z );
		mat4 boneMatW = getBoneMatrix( skinIndex.w );
	#endif
	//==============================================
	//==============================================
	//==============================================

	//==============================================
	// THREE.ShaderChunk[ "skinnormal_vertex" ],
	//==============================================
	#ifdef USE_SKINNING
		mat4 skinMatrix = mat4( 0.0 );
		skinMatrix += skinWeight.x * boneMatX;
		skinMatrix += skinWeight.y * boneMatY;
		skinMatrix += skinWeight.z * boneMatZ;
		skinMatrix += skinWeight.w * boneMatW;
		skinMatrix  = bindMatrixInverse * skinMatrix * bindMatrix;

		#ifdef USE_MORPHNORMALS
		vec4 skinnedNormal = skinMatrix * vec4( morphedNormal, 0.0 );
		#else
		vec4 skinnedNormal = skinMatrix * vec4( normal, 0.0 );
		#endif
	#endif
	//==============================================
	//==============================================
	//==============================================

	// normal, tangent and binormal vectors

	#ifdef USE_SKINNING
		vNormal = normalize( normalMatrix * skinnedNormal.xyz );

		vec4 skinnedTangent = skinMatrix * vec4( tangent.xyz, 0.0 );
		vTangent = normalize( normalMatrix * skinnedTangent.xyz );
	#else
		vNormal = normalize( normalMatrix * normal );
		vTangent = normalize( normalMatrix * tangent.xyz );
	#endif

	vBinormal = normalize( cross( vNormal, vTangent ) * tangent.w );

	vUv = uv * uRepeat + uOffset;

	// displacement mapping

	vec3 displacedPosition;

	#ifdef VERTEX_TEXTURES

		if ( enableDisplacement ) {

			vec3 dv = texture2D( tDisplacement, uv ).xyz;
			float df = uDisplacementScale * dv.x + uDisplacementBias;
			displacedPosition = position + normalize( normal ) * df;

		} else {

			#ifdef USE_SKINNING

				vec4 skinVertex = bindMatrix * vec4( position, 1.0 );

				vec4 skinned = vec4( 0.0 );
				skinned += boneMatX * skinVertex * skinWeight.x;
				skinned += boneMatY * skinVertex * skinWeight.y;
				skinned += boneMatZ * skinVertex * skinWeight.z;
				skinned += boneMatW * skinVertex * skinWeight.w;
				skinned  = bindMatrixInverse * skinned;

				displacedPosition = skinned.xyz;

			#else

				displacedPosition = position;

			#endif

		}

	#else

		#ifdef USE_SKINNING

			vec4 skinVertex = bindMatrix * vec4( position, 1.0 );

			vec4 skinned = vec4( 0.0 );
			skinned += boneMatX * skinVertex * skinWeight.x;
			skinned += boneMatY * skinVertex * skinWeight.y;
			skinned += boneMatZ * skinVertex * skinWeight.z;
			skinned += boneMatW * skinVertex * skinWeight.w;
			skinned  = bindMatrixInverse * skinned;

			displacedPosition = skinned.xyz;

		#else

			displacedPosition = position;

		#endif

	#endif

	//

	vec4 mvPosition = modelViewMatrix * vec4( displacedPosition, 1.0 );
	vec4 worldPosition = modelMatrix * vec4( displacedPosition, 1.0 );

	gl_Position = projectionMatrix * mvPosition;

	//===========================================
	// THREE.ShaderChunk[ "logdepthbuf_vertex" ],
	//===========================================
	#ifdef USE_LOGDEPTHBUF
		gl_Position.z = log2(max(1e-6, gl_Position.w + 1.0)) * logDepthBufFC;
	
		#ifdef USE_LOGDEPTHBUF_EXT
			vFragDepth = 1.0 + gl_Position.w;
		#else
			gl_Position.z = (gl_Position.z - 1.0) * gl_Position.w;
		#endif
	#endif
	//===========================================
	//===========================================
	//===========================================

	vWorldPosition = worldPosition.xyz;
	vViewPosition = -mvPosition.xyz;

	// shadows

	#ifdef USE_SHADOWMAP
		for( int i = 0; i < MAX_SHADOWS; i ++ ) {
			vShadowCoord[ i ] = shadowMatrix[ i ] * worldPosition;
		}
	#endif

}
